/*
 * USB220_GPIB driver - 1.2
 *
 * Copyright (C) 2009 Ratoc Systems, Inc.
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License as
 *	published by the Free Software Foundation, version 2.
 *
 * This driver is based on the 2.6.3 version of drivers/usb/usb-skeleton.c 
 * but has been rewritten to be easy to read and use, as no locks are now
 * needed anymore.
 *
 * Changelog:
 * 2005/09/01  start coding
 * 2006/12/13  changed  interruptible_sleep_on()->wait_event_interruptible()
 * 2009/12/02  Fix function gpib_write and gpib_read.
 *
 **/

#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/kref.h>
#include <asm/uaccess.h>
#include <linux/usb.h>
#include <linux/wait.h>

#include "usb220_gpib.h"

/* Define these values to match your devices */
#define USB_GPIB_VENDOR_ID	0x0584
#define USB_GPIB_PRODUCT_ID	0xA220


/* table of devices that work with this driver */
static struct usb_device_id gpib_table [] = {
	{ USB_DEVICE(USB_GPIB_VENDOR_ID, USB_GPIB_PRODUCT_ID) },
	{ }					/* Terminating entry */
};
MODULE_DEVICE_TABLE (usb, gpib_table);


/* Get a minor range for your devices from the usb maintainer */
#define USB_GPIB_MINOR_BASE	208
#define USBGPIB_TIMEOUT	(2*HZ)			/* 2 seconds */

/* Structure to hold all of our device specific stuff */
struct usb_gpib {
	struct usb_device *udev;			/* the usb device for this device */
	//struct usb_interface *interface;		/* the interface for this device */
	unsigned char ctrlBuf[64];
	__u8			bulk_in_endpointAddr;	/* the address of the bulk in endpoint = 0x86 (pipe1) */
	__u8			bulk_out_endpointAddr;	/* the address of the bulk out endpoint = 0x02 (pipe0) */
	__u8			interrupt_endpointAddr;	/* the address of the interrupt endpoint = 0x81 (pipe2) */

	void __user *readBuf;
	wait_queue_head_t wait_bulk_out;
	wait_queue_head_t wait_bulk_in;
	int        wait_bulk_out_complete;
	int        wait_bulk_in_complete;

	struct urb *intUrb;		/* urb for interrupt */
	unsigned char intBuf[16];
	int        bInterrupting;
	wait_queue_head_t wait_interrupt;
	int        wait_interrupt_complete;
	struct semaphore lock;          /* general race avoidance */

	struct kref		kref;
};
#define to_gpib_dev(d) container_of(d, struct usb_gpib, kref)

static struct usb_driver gpib_driver;


static void gpib_delete(struct kref *kref)
{	
	struct usb_gpib *dev = to_gpib_dev(kref);

	usb_put_dev(dev->udev);
	kfree (dev);
}

static int gpib_open(struct inode *inode, struct file *file)
{
	struct usb_gpib *dev;
	struct usb_interface *interface;
	int subminor;
	int retval = 0;

	subminor = iminor(inode);
	info("gpib_open subminor = %d", subminor);

	interface = usb_find_interface(&gpib_driver, subminor);
	if (!interface) {
		err ("%s - error, can't find device for minor %d",
		     __FUNCTION__, subminor);
		retval = -ENODEV;
		goto exit;
	}

	dev = usb_get_intfdata(interface);
	if (!dev) {
		retval = -ENODEV;
		goto exit;
	}

	/* increment our usage count for the device */
	kref_get(&dev->kref);

	/* save our object in the file's private structure */
	file->private_data = dev;

exit:
	return retval;
}

static int gpib_release(struct inode *inode, struct file *file)
{
	struct usb_gpib *dev;

	info("gpib_release subminor = %d", iminor(inode));
	dev = (struct usb_gpib *)file->private_data;
	if (dev == NULL)
		return -ENODEV;

	/* decrement the count on our device */
	kref_put(&dev->kref, gpib_delete);
	return 0;
}

static void gpib_read_bulk_callback(struct urb *urb, struct pt_regs *regs)
{
	struct usb_gpib *dev = urb->context;
	dev->wait_bulk_in_complete = 1;
	wake_up_interruptible(&dev->wait_bulk_in);
}

static ssize_t gpib_read(struct file *file, char *buffer, size_t count, loff_t *ppos)
{
	struct usb_gpib *dev;
	int retval = 0;
	struct urb *urb = NULL;
	char *buf = NULL;

	dev = (struct usb_gpib *)file->private_data;

	down(&(dev->lock));

	/* verify that we actually have some data to write */
	if (count == 0) {
		goto exit;
	}

	dev->readBuf = (void __user *) buffer;

	/* create a urb, and a buffer for it, and copy the data to the urb */
	urb = usb_alloc_urb(0, GFP_KERNEL);
	if (!urb) {
		retval = -ENOMEM;
		goto exit;
	}

	buf = usb_buffer_alloc(dev->udev, count, GFP_KERNEL, &urb->transfer_dma);
	if (!buf) {
		retval = -ENOMEM;
		goto exit;
	}

	/* initialize the urb properly */
	usb_fill_bulk_urb(urb, dev->udev,
			  usb_rcvbulkpipe(dev->udev, dev->bulk_in_endpointAddr),
			  buf, count, gpib_read_bulk_callback, dev);
	urb->transfer_flags |= URB_NO_TRANSFER_DMA_MAP;

	/* send the data out the bulk port */
	dev->wait_bulk_in_complete = 0;
	retval = usb_submit_urb(urb, GFP_KERNEL);
	if (retval) {
		info("%s - failed submitting read urb, error %d", __FUNCTION__, retval);
		goto exit;
	}
	//wait_event_interruptible(dev->wait_bulk_in, (dev->wait_bulk_complete) );
	retval = wait_event_interruptible_timeout(dev->wait_bulk_in, (dev->wait_bulk_in_complete), HZ*10);
	if (retval <= 0) {
		retval = -1; // ioctl error
		goto exit;
	}

	if (urb->status) {
		info("%s - nonzero read bulk status received: %d",
				__FUNCTION__, urb->status);
		usb_clear_halt(dev->udev, usb_rcvbulkpipe(dev->udev, dev->bulk_in_endpointAddr));
		retval = 0;
		goto exit;
	}else{
		if (copy_to_user(dev->readBuf, buf, count)) {
			retval = -EFAULT;
			goto exit;
		}
		retval = urb->actual_length;
	}

exit:
	/* release our reference to this urb, the USB core will eventually free it entirely */
	usb_buffer_free(dev->udev, count, buf, urb->transfer_dma);
	usb_free_urb(urb);
	up(&(dev->lock));
	return retval;
}

static void gpib_write_bulk_callback(struct urb *urb, struct pt_regs *regs)
{
	struct usb_gpib *dev = urb->context;
	dev->wait_bulk_out_complete = 1;
	wake_up_interruptible(&dev->wait_bulk_out);
}

static ssize_t gpib_write(struct file *file, const char *user_buffer, size_t count, loff_t *ppos)
{
	struct usb_gpib *dev;
	int retval = 0;
	struct urb *urb = NULL;
	char *buf = NULL;

	dev = (struct usb_gpib *)file->private_data;

	down(&(dev->lock));

	/* verify that we actually have some data to write */
	if (count == 0) {
		goto exit;
	}

	/* create a urb, and a buffer for it, and copy the data to the urb */
	urb = usb_alloc_urb(0, GFP_KERNEL);
	if (!urb) {
		retval = -ENOMEM;
		goto exit;
	}

	buf = usb_buffer_alloc(dev->udev, count, GFP_KERNEL, &urb->transfer_dma);
	if (!buf) {
		retval = -ENOMEM;
		goto exit;
	}

	if (copy_from_user(buf, user_buffer, count)) {
		retval = -EFAULT;
		goto exit;
	}

	/* initialize the urb properly */
	usb_fill_bulk_urb(urb, dev->udev,
			  usb_sndbulkpipe(dev->udev, dev->bulk_out_endpointAddr),
			  buf, count, gpib_write_bulk_callback, dev);
	urb->transfer_flags |= URB_NO_TRANSFER_DMA_MAP;

	/* send the data out the bulk port */
	dev->wait_bulk_out_complete = 0;
	retval = usb_submit_urb(urb, GFP_KERNEL);
	if (retval) {
		info("%s - failed submitting write urb, error %d", __FUNCTION__, retval);
		goto exit;
	}
	//wait_event_interruptible(dev->wait_bulk, (dev->wait_bulk_complete) );
	retval = wait_event_interruptible_timeout(dev->wait_bulk_out, (dev->wait_bulk_out_complete), HZ*10);
	if (retval <= 0) {
		retval = -1; // ioctl error
		goto exit;
	}

	/* sync/async unlink faults aren't errors */
	if (urb->status) {
		info("%s - nonzero write bulk status received: %d",
		    __FUNCTION__, urb->status);
		usb_clear_halt(dev->udev, usb_sndbulkpipe(dev->udev, dev->bulk_out_endpointAddr));
		retval = 0;
		goto exit;
	}else{
		retval = urb->actual_length;
	}

exit:
	/* release our reference to this urb, the USB core will eventually free it entirely */
	usb_buffer_free(dev->udev, count, buf, urb->transfer_dma);
	usb_free_urb(urb);
	up(&(dev->lock));
	return retval;
}

static void gpib_interrupt_callback(struct urb *urb, struct pt_regs *regs)
{
	struct usb_gpib *dev = urb->context;
	info("gpib_interrupt_callback->wake_up");
	dev->wait_interrupt_complete = 1;
	wake_up_interruptible(&dev->wait_interrupt);
}

static int
gpib_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	struct GpibCommand gpib_cmd;
	struct usb_gpib *dev;
	void __user *data;
	unsigned int	EndPoint;
	int result, retries;
	int retval = 0;

	dev = (struct usb_gpib *)file->private_data;

	down(&(dev->lock));

	/* Sanity check to make sure rio is connected, powered, etc */
    if ( dev == NULL || dev->udev == NULL )
	{
		retval = -ENODEV;
		goto err_out;
	}

	switch (cmd) {
	case GPIB_RECV_COMMAND:
		data = (void __user *) arg;
		if (data == NULL)
			break;
		if (copy_from_user(&gpib_cmd, data, sizeof(struct GpibCommand))) {
			retval = -EFAULT;
			goto err_out;
		}
		if (gpib_cmd.length < 0 || gpib_cmd.length > PAGE_SIZE) {
			retval = -EINVAL;
			goto err_out;
		}

		//requesttype = gpib_cmd.requesttype | USB_DIR_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE;
		//info
		//    ("sending command:reqtype=%0x req=%0x value=%0x index=%0x len=%0x",
		//     gpib_cmd.requesttype, gpib_cmd.request, gpib_cmd.value,
		//     gpib_cmd.index, gpib_cmd.length);
		/* Send rio control message */
		retries = 3;
		while (retries) {
			result = usb_control_msg(dev->udev,
						 usb_rcvctrlpipe(dev->udev, 0),
						 gpib_cmd.request,
						 gpib_cmd.requesttype,
						 gpib_cmd.value,
						 gpib_cmd.index, dev->ctrlBuf,
						 gpib_cmd.length,
						 gpib_cmd.timeout);
			if (result == -ETIMEDOUT){
				info("Error ETIMEDOUT ioctrl. retries = %d", retries);
				retries--;
			}else if (result < 0) {
				info("Error executing ioctrl. code = %d", result);
				retries = 0;
			} else {
				dbg("Executed ioctl. Result = %d (buffer=%02x)",
				     result, dev->ctrlBuf[0]);
				if (copy_to_user(gpib_cmd.buffer, dev->ctrlBuf,
						 gpib_cmd.length)) {
					retval = -EFAULT;
					goto err_out;
				}
				retries = 0;
			}

			/* gpib_cmd.buffer contains a raw stream of single byte
			   data which has been returned from rio.  Data is
			   interpreted at application level.  For data that
			   will be cast to data types longer than 1 byte, data
			   will be little_endian and will potentially need to
			   be swapped at the app level */

		}

		gpib_cmd.length = result;
		if (copy_to_user(data, &gpib_cmd, sizeof(struct GpibCommand))) {
			retval = -EFAULT;
			goto err_out;
		}
		//info("gp_recv  req = 0x%x, length = 0x%x, (buf=0x%x), retval = 0x%x", gpib_cmd.request, gpib_cmd.length, dev->ctrlBuf[0], retval);
		break;

	case GPIB_SEND_COMMAND:
		data = (void __user *) arg;
		if (data == NULL)
			break;
		if (copy_from_user(&gpib_cmd, data, sizeof(struct GpibCommand))) {
			retval = -EFAULT;
			goto err_out;
		}
		if (gpib_cmd.length < 0 || gpib_cmd.length > PAGE_SIZE) {
			retval = -EINVAL;
			goto err_out;
		}
		if (copy_from_user(dev->ctrlBuf, gpib_cmd.buffer, gpib_cmd.length)) {
			retval = -EFAULT;
			goto err_out;
		}

		//requesttype = rio_cmd.requesttype | USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE;
		//info("sending command: reqtype=%0x req=%0x value=%0x index=%0x len=%0x",
		//     gpib_cmd.requesttype, gpib_cmd.request, gpib_cmd.value,
		//     gpib_cmd.index, gpib_cmd.length);
		/* Send rio control message */
		retries = 3;
		while (retries) {
			result = usb_control_msg(dev->udev,
						 usb_sndctrlpipe(dev->udev, 0),
						 gpib_cmd.request,
						 gpib_cmd.requesttype,
						 gpib_cmd.value,
						 gpib_cmd.index, dev->ctrlBuf,
						 gpib_cmd.length,
						 gpib_cmd.timeout);
			if (result == -ETIMEDOUT){
				info("ETIMEDOUT ioctrl. retries = %d", retries);
				retries--;
			}else if (result < 0) {
				info("Error executing ioctrl. code = %d", result);
				retries = 0;
			} else {
				dbg("Executed ioctl. Result = %d", result);
				retries = 0;
			}

		}

		gpib_cmd.length = result;
		if (copy_to_user(data, &gpib_cmd, sizeof(struct GpibCommand))) {
			retval = -EFAULT;
			goto err_out;
		}
		//info("gp_send   req = 0x%x, length = 0x%x, retval = 0x%x", gpib_cmd.request, gpib_cmd.length, retval);
		break;

	case GPIB_START_INTERRUPT:
		data = (void __user *) arg;
		if (data == NULL)
			break;

		dev->wait_interrupt_complete = 0;
		if (usb_submit_urb(dev->intUrb, GFP_KERNEL)) {
			info("usb_submit_urb(int irq) failed");
			usb_kill_urb(dev->intUrb);
			retval = -EFAULT;
			goto err_out;
		}

		dev->bInterrupting = 1;
		info("gpib_ioctl INTERRUPT...->sleep_on");
		up(&(dev->lock));
		wait_event_interruptible(dev->wait_interrupt, (dev->wait_interrupt_complete) );
		down(&(dev->lock));
		info("gpib_ioctl INTERRUPT...sleep_on->");
		dev->bInterrupting = 0;

		if (copy_to_user(data, dev->intBuf, sizeof(unsigned char))) {
			retval = -EFAULT;
			goto err_out;
		}
		info("gp_int   (buf[0]=0x%x), retval = 0x%x", dev->intBuf[0], retval);
		break;

	case GPIB_STOP_INTERRUPT:
		if(dev->bInterrupting){
			info("gpib_ioctl GPIB_STOP_INTERRUPT usb_kill_urb");
			usb_kill_urb(dev->intUrb);
		}else{
			info("gpib_ioctl GPIB_STOP_INTERRUPT ...");
		}
		break;

	case GPIB_ABORT_PIPE:
		data = (void __user *) arg;
		if (data == NULL)
			break;
		if (copy_from_user(&EndPoint, data, sizeof(int))) {
			retval = -EFAULT;
			goto err_out;
		}
		if(EndPoint == 0/*PIPE_FOR_BULKOUT*/){
			info("gpib_ioctl GPIB_ABORT_PIPE BULKOUT ...");
		}
		if(EndPoint == 1/*PIPE_FOR_BULKIN*/){
			info("gpib_ioctl GPIB_ABORT_PIPE PIPE_FOR_BULKIN ...");
		}
		if(EndPoint == 2/*PIPE_FOR_INTERRUPT*/){
			info("gpib_ioctl GPIB_ABORT_PIPE PIPE_FOR_INTERRUPT usb_kill_urb");
			usb_kill_urb(dev->intUrb);
		}
		break;

	default:
		retval = -ENOTTY;
		break;
	}

err_out:
	up(&(dev->lock));
	return retval;
}

static struct file_operations gpib_fops = {
	.owner =	THIS_MODULE,
	.read =		gpib_read,
	.write =	gpib_write,
	.ioctl =	gpib_ioctl,
	.open =		gpib_open,
	.release =	gpib_release,
};

/* 
 * usb class driver info in order to get a minor number from the usb core,
 * and to have the device registered with devfs and the driver core
 */
static struct usb_class_driver gpib_class = {
	.name =		"usb/gpib%d",
	.fops =		&gpib_fops,
	.minor_base =	USB_GPIB_MINOR_BASE,
};

static int gpib_probe(struct usb_interface *interface, const struct usb_device_id *id)
{
	struct usb_gpib *dev = NULL;
	struct usb_host_interface *iface_desc;
	struct usb_endpoint_descriptor *endpoint;
	int i;
	int retval = -ENOMEM;

	/* allocate memory for our device state and initialize it */
	dev = kmalloc(sizeof(*dev), GFP_KERNEL);
	if (dev == NULL) {
		err("Out of memory");
		goto error;
	}
	memset(dev, 0x00, sizeof(*dev));
	kref_init(&dev->kref);

	dev->udev = usb_get_dev(interface_to_usbdev(interface));

	/* set up the endpoint information */
	/* use only the first bulk-in and bulk-out endpoints */
	iface_desc = interface->cur_altsetting;
	for (i = 0; i < iface_desc->desc.bNumEndpoints; ++i) {
		endpoint = &iface_desc->endpoint[i].desc;

		if (!dev->bulk_in_endpointAddr &&
		    (endpoint->bEndpointAddress & USB_DIR_IN) &&
		    ((endpoint->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK)
					== USB_ENDPOINT_XFER_BULK)) {
			/* we found a bulk in endpoint */
			dev->bulk_in_endpointAddr = endpoint->bEndpointAddress;
		}

		if (!dev->bulk_out_endpointAddr &&
		    !(endpoint->bEndpointAddress & USB_DIR_IN) &&
		    ((endpoint->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK)
					== USB_ENDPOINT_XFER_BULK)) {
			/* we found a bulk out endpoint */
			dev->bulk_out_endpointAddr = endpoint->bEndpointAddress;
		}

		if (!dev->interrupt_endpointAddr &&
		    ((endpoint->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK)
					== USB_ENDPOINT_XFER_INT)) {
			/* we found a interrupt endpoint */
			dev->interrupt_endpointAddr = endpoint->bEndpointAddress;
			dev->intUrb = usb_alloc_urb(0, GFP_KERNEL);
			usb_fill_int_urb(dev->intUrb, dev->udev,
					usb_rcvintpipe(dev->udev, endpoint->bEndpointAddress),
					dev->intBuf, sizeof(unsigned char), gpib_interrupt_callback, dev, endpoint->bInterval/* 1ms */);
		}
	}
	if (!(dev->bulk_in_endpointAddr && dev->bulk_out_endpointAddr)) {
		err("Could not find both bulk-in and bulk-out endpoints");
		goto error;
	}
	if (!(dev->interrupt_endpointAddr && dev->intUrb)) {
		err("Could not find interrupt endpoints");
		goto error;
	}

	/* save our data pointer in this interface device */
	usb_set_intfdata(interface, dev);

	/* we can register the device now, as it is ready */
	retval = usb_register_dev(interface, &gpib_class);
	if (retval) {
		/* something prevented us from registering this driver */
		err("Not able to get a minor for this device.");
		usb_set_intfdata(interface, NULL);
		goto error;
	}

	/* let the user know what node this device is now attached to */
	info("USB GPIB device now attached... interface->minor = %d", interface->minor);
	init_waitqueue_head(&dev->wait_bulk_out);
	init_waitqueue_head(&dev->wait_bulk_in);
	init_waitqueue_head(&dev->wait_interrupt);
	dev->bInterrupting = 0;
	init_MUTEX(&(dev->lock));
	return 0;

error:
	if (dev)
		kref_put(&dev->kref, gpib_delete);
	return retval;
}

static void gpib_disconnect(struct usb_interface *interface)
{
	struct usb_gpib *dev;
	int minor = interface->minor;

	/* prevent gpib_open() from racing skel_disconnect() */
	lock_kernel();

	dev = usb_get_intfdata(interface);

	usb_kill_urb(dev->intUrb);
	usb_free_urb(dev->intUrb);

	usb_set_intfdata(interface, NULL);

	/* give back our minor */
	usb_deregister_dev(interface, &gpib_class);

	unlock_kernel();

	/* decrement our usage count */
	kref_put(&dev->kref, gpib_delete);

	info("USB GPIB #%d now disconnected", minor);
}

static struct usb_driver gpib_driver = {
	.name =		"ratocgpib",
	.probe =	gpib_probe,
	.disconnect =	gpib_disconnect,
	.id_table =	gpib_table,
};

static int __init usb_gpib_init(void)
{
	int result;

	/* register this driver with the USB subsystem */
	result = usb_register(&gpib_driver);
	if (result)
		err("usb_register failed. Error number %d", result);

	return result;
}

static void __exit usb_gpib_exit(void)
{
	/* deregister this driver with the USB subsystem */
	usb_deregister(&gpib_driver);
}

module_init (usb_gpib_init);
module_exit (usb_gpib_exit);

MODULE_LICENSE("GPL");
